package clustering;

import java.io.*;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.cookie.CookiePolicy;
import org.apache.commons.httpclient.methods.*;
import org.apache.commons.httpclient.methods.multipart.*;

/**
 * <p><code>MediaWikiComm</code> is a class that provides the abilility to
 * load past settings from the <code>DemoApplet</code>'s wiki page upon
 * applet startup.  It also provides the ability to save settings to the page.
 * </p>
 * 
 * @author R. Scarberry
 */
public class MediaWikiComm {

    // Names of form data elements in the HTML for the 
    // MediaWiki edit tab. (Elements of the form element "editform".)
    private static final String WP_SECTION = "wpSection";
    private static final String WP_STARTTIME = "wpStarttime";
    private static final String WP_EDITTIME = "wpEdittime";
    private static final String WP_SCROLLTOP = "wpScrolltop";
    private static final String WP_TEXTBOX1 = "wpTextbox1";
    private static final String WP_SUMMARY = "wpSummary";
    private static final String WP_SAVE = "wpSave";
    private static final String WP_EDITTOKEN = "wpEditToken";
    private static final String WP_AUTOSUMMARY = "wpAutoSummary";
    
    // The page URL in the format http://<hostname>/mediawiki/index.php?title=<pagetitle>
    private String mPageURL;
    // Tags used for the applet (typically "clustering") and the applet data
    // (typically "clustering_data").
    private String mExtensionTag, mDataTag;
    // The method used to embed the applet -- tag, function, or template.
    private int mExtensionType = ClusteringDemoSettingsFactory.EXTENSION_TYPE_TAG;
    // Cookies posted in the header of every request.  Even-numbered elements are the
    // cookie names, odd-numbered are the values.  These are constructed from
    // the UserName, UserID, _session, and CookiePrefix parameters to the applet.
    private String[] mCookiePairs;
	
    // The contents of the form with the name/id "editform".  Set by loadData(), several
    // elements must be posted back to the server by saveData().
    private EditFormData mPageData;
	
    // Object for communication with the server via HTTP, reused in
    // every request.
    private HttpClient mHttpClient = new HttpClient();

    /**
     * Constructor
     * 
     * @param pageURL       the page URL
     * @param extensionTag  tag used to embed the applet in the unrendered text of the page
     * @param dataTag       tag used to embed the applet data in the unrendered text of the page
     * @param extensionType how the applet was embedded (tag, function, or template)
     * @param cookiePairs   cookies to be included in every server request
     */
    public MediaWikiComm(String pageURL, 
            String extensionTag, String dataTag, String extensionType,
            String[] cookiePairs) {

        // Ensure required parameters are set.
        if (pageURL == null || extensionTag == null || dataTag == null || extensionType == null) {
            throw new NullPointerException();
        }
        
        mPageURL = pageURL;
        mExtensionTag = extensionTag;
        mDataTag = dataTag;

        // Translate the extension type into the appropriate integer version.
        if (extensionType.equalsIgnoreCase("tag")) {
            mExtensionType = ClusteringDemoSettingsFactory.EXTENSION_TYPE_TAG;
        } else if (extensionType.equalsIgnoreCase("function")) {
            mExtensionType = ClusteringDemoSettingsFactory.EXTENSION_TYPE_FUNCTION;
        } else if (extensionType.equalsIgnoreCase("template")) {
            mExtensionType = ClusteringDemoSettingsFactory.EXTENSION_TYPE_TEMPLATE;
        }

        // Make a defensive copy of the cookie pairs.
        mCookiePairs = cookiePairs != null ? ((String[]) cookiePairs.clone()) : new String[0];
    }
	
    /**
     * Loads the data already on the wiki page and returns it as an
     * array of <code>ClusteringDemoSetting</code>s objects.
     * 
     * @return an array of objects, possibly of length 0, but never null.
     * 
     * @throws IOException
     */
    public ClusteringDemoSettings[] loadData() throws IOException {
    	mPageData = loadEditFormData();
    	return ClusteringDemoSettingsFactory.parseFromContent(
                mPageData.getEditText(), mDataTag);
    }
    
    /**
     * Method which loads downloads the data from the wiki page.
     * 
     * @return an instance of EditFormData, which encapsulates the 
     *   important elements of the page's editform.
     *   
     * @throws IOException
     */
    private EditFormData loadEditFormData() throws IOException {

        // Get the HTML from the page's edit tab.
        GetMethod method = new GetMethod(mPageURL + "&action=edit");
        int responseCode = 200;
        String responseBody = null;

        // Tell the HttpClient library not to worry about the cookies.  We'll set
        // them manually.
        method.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        // Get a string representation of the cookies.
        String cookies = getCookieString();
        if (cookies.length() > 0) {
            // Put the cookies in the request header.
            method.setRequestHeader("Cookie", cookies);
        }

        try {

            // Perform the get.
            responseCode = mHttpClient.executeMethod(method);

            // Returns 200 on success.  But techically, anything in the 200s 
            // is supposed to mean success.
            if (responseCode >= 200 && responseCode < 300) {

                // Read the entire response and place it into responseBody.
                // This will be the HTML for the entire page.  (The same as
                // what you see when you view source on the edit tab in the browser.)
                
                String charSet = method.getResponseCharSet();

                // method.getResponseBody() returns everything as byte[].
                // method.getResponseBodyAsString() returns everything as a String, but
                // the charSet might be ignored.  So we'll use method.getResponseBodyAsStream() and
                // read it line-by-line.
                BufferedReader br = new BufferedReader(new InputStreamReader(
                        method.getResponseBodyAsStream(), charSet));

                StringWriter sw = new StringWriter();
                PrintWriter pw = new PrintWriter(sw);
                String line = null;
                while((line = br.readLine()) != null) {
                    pw.println(line);
                }
                pw.flush();

                responseBody = sw.toString();

            } else { // Bad response code -- return an IOException with the status.

                throw new IOException(method.getStatusText() + " (" + responseCode + ")");

            }

        } finally {
            
            // Always release the connect when done.
            method.releaseConnection();

        }

        // To have fallen through, the response code had to have been ok.
        // Now parse out the editform into the utility class HtmlForm.
        HtmlForm form = HtmlUtilities.extractForm("editform", responseBody);
        String empty = "";
        String startTime = empty, editTime = empty, 
          editText = empty, autoSummary = empty, editToken = empty;

        if (form != null) {

            int sz = form.getElementCount();

            // Step through the form elements to find key items needed to
            // construct an instance of EditFormData.
            for (int i=0; i<sz; i++) {

                HtmlForm.Element element = form.getElement(i);
                HtmlForm.ElementAttribute nameAttr = element.getElementAttributeByName("name");
                HtmlForm.ElementAttribute valueAttr = element.getElementAttributeByName("value");

                String name = nameAttr != null ? nameAttr.getValue() : null;

                try {
                   
                    if (WP_STARTTIME.equals(name)) {
                        startTime = valueAttr.getValue();
                    } else if (WP_EDITTIME.equals(name)) {
                        editTime = valueAttr.getValue();
                    } else if (WP_TEXTBOX1.equals(name)) {
                        // This item contains what you see in the text area of the edit tab of the browser.
                        // The reason this is read line-by-line is to translate carriage-return/linefeeds.
                        BufferedReader br2 = new BufferedReader(new StringReader(element.getEmbeddedText()));
                        StringWriter sw2 = new StringWriter();
                        PrintWriter pw2 = new PrintWriter(sw2);
                        String line2 = null;
                        while((line2 = br2.readLine()) != null) {
                            pw2.println(line2);
                        }
                        pw2.close();
                        try { br2.close(); } catch (IOException wontHappen) {}
                        // Set editText.
                        editText = sw2.getBuffer().toString();
                    } else if (WP_AUTOSUMMARY.equals(name)) {
                        autoSummary = valueAttr.getValue();
                    } else if (WP_EDITTOKEN.equals(name)) {
                        // The edit token must be posted back to the server with saves
                        // for them to work.  The server maps the edit token to the session
                        // in the request cookie.
                        editToken = valueAttr.getValue();
                    }
                } catch (RuntimeException rte) {
                    rte.printStackTrace();
                    // Just in case invalid form data results in a 
                    // NullPointerException, IllegalArgumentException, etc...
                    throw new IOException("invalid editform data: " + rte.toString());
                }
            } // for (int i...

        } else { // form == null

            throw new IOException("no edit form on page");

        }

        // Return the data in an instance of EditFormData.
        return new EditFormData(startTime, editTime, editText, autoSummary, editToken);
    }

    /**
     * Saves data back to the wiki page of the clustering demo applet.
     * 
     * @param settings an array of <code>ClusteringDemoSetting</code>s object, 
     *   which are converted into a block of XML and embedded on the page.
     *   
     * @throws IOException
     */
    public void saveData(ClusteringDemoSettings[] settings) throws IOException {

        // Returns a new version of the edit text with the <clustering_data> block
        // replaced or added.
    	String newEditText = ClusteringDemoSettingsFactory.updateContent(
                mPageData.getEditText(), 
    		mExtensionTag, mDataTag, mExtensionType, settings);
        
        if (newEditText == null) {
            // Bail out if the new data could not be generated.
            throw new IOException("could not generate updated page content");
        }
        
        // Now post the data as multipart form data.
    	PostMethod method = new PostMethod(mPageURL + "&action=submit");
        int responseCode = 200;
        
        // Add the cookies to the request header manually.
        method.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        String cookies = getCookieString();
        if (cookies.length() > 0) {
            method.setRequestHeader("Cookie", cookies);
        }
        
        // Construct the parts for the request body from the items in mPageData
        // downloaded earlier and the new edit text.
        Part[] parts = {
                new StringPart(WP_SECTION, ""),
                new StringPart(WP_STARTTIME, mPageData.getStartTime()),
                new StringPart(WP_EDITTIME, mPageData.getEditTime()),
                new StringPart(WP_SCROLLTOP, ""),
                new StringPart(WP_TEXTBOX1, newEditText),
                new StringPart(WP_SUMMARY, ""),
                new StringPart(WP_SAVE, "Save page"),
                new StringPart(WP_EDITTOKEN, mPageData.getEditToken()),
                new StringPart(WP_AUTOSUMMARY, mPageData.getAutoSummary())
        };
        
        // Add the parts to the request.
        method.setRequestEntity(new MultipartRequestEntity(parts, method.getParams()));

        try {

            // Execute the request.
            responseCode = mHttpClient.executeMethod(method);
            
            // This time, 302 means success. 302 is returned when the server
            // attempts a page redirection.
            // 200 means the save failed probably because another user posted some
            // edits since loadData() was called.
            if (responseCode >= 200 && responseCode < 300) {
                // The calling code will respond by reloading the data, merging, and
                // attempting to save again.
                throw new ConcurrentEditException();
            } else if (responseCode != 302) {
                // Some other error code.
                throw new IOException(method.getStatusText() + " (" + responseCode + ")");
            }
            
        } finally {
            
            // Always release the connection.
            method.releaseConnection();
        }
	
    }
        	
    /**
     * Returns a cookie string suitable for placing in the 
     * Cookie request header.
     * 
     * @return
     */
    private String getCookieString() {
        StringBuffer cookieBuf = new StringBuffer();
        int pairs = mCookiePairs != null ? mCookiePairs.length/2 : 0;
        for (int i=0; i<pairs; i+=2) {
            if (i > 0) cookieBuf.append("; ");
            cookieBuf.append(mCookiePairs[i]);
            cookieBuf.append('=');
            cookieBuf.append(mCookiePairs[i+1]);
        }
        return cookieBuf.toString();
    }
    

    /**
     * Object which encapsulates data parsed from the edit tab page.  
     * This data must be posted to the server in order to successfully 
     * post edits.
     */
    static class EditFormData {
     
        // The MediaWiki server uses these to 
        // determine if another user made edits since this data 
        // was downloaded.
        private String mStartTime, mEditTime;
        // The contents of the text area on the edit tab.  Contains the
        // text used to embed the applet and the applet data.
        private String mEditText;
        private String mAutoSummary;
        // Token which must be passed to the server to save edits.
        private String mEditToken;
        
        /**
         * Constructor.
         * 
         * @param startTime
         * @param editTime
         * @param editText
         * @param autoSummary
         * @param editToken
         */
        EditFormData(String startTime, String editTime, String editText, 
                String autoSummary, String editToken) {
            this.mStartTime = startTime;
            this.mEditTime = editTime;
            this.mEditText = editText;
            this.mAutoSummary = autoSummary;
            this.mEditToken = editToken;
        }
        
        // Getter methods.
        //
        String getStartTime() { return mStartTime; }
        String getEditTime() { return mEditTime; }
        String getEditText() { return mEditText; }
        String getAutoSummary() { return mAutoSummary; }
        String getEditToken() { return mEditToken; }
    } 
    
    /**
     * Special case of IOException to be thrown when saving data to the server fails because
     * another user saved data.
     */
    public static class ConcurrentEditException extends IOException {
    
    	public ConcurrentEditException(String message) {
    		super(message);
    	}
    	
    	public ConcurrentEditException() {
    	}
    	
    }
}
