package clustering;

import java.awt.*;
import javax.swing.*;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYItemRenderer;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;

import java.awt.event.*;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.*;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Random;

/**
 * The main panel of the clustering demo user interface.
 * 
 * @author R. Scarberry
 *
 */
public class ClusteringDemoPanel extends JPanel 
implements ClusteringListener {

	// Defaults settings
    public static final int DEFAULT_COORDS 			= 1000;
    public static final int DEFAULT_CLUSTERS 		= 10;
    public static final long DEFAULT_SEED 			= 1234L;
    public static final double DEFAULT_STANDARD_DEV = 0.02;
    
    // The various UI components.
    private JPanel mCenterPanel = new JPanel();
    private JPanel mNorthPanel = new JPanel();
    private JPanel mPrevRunPanel = new JPanel();
    private JLabel mCoordLabel = new JLabel();
    private JLabel mClusterLabel = new JLabel();
    private JLabel mRandomLabel = new JLabel();
    private JLabel mStandardDevLabel = new JLabel();
    private JTextField mCoordTF = new JTextField();
    private JTextField mClustersTF = new JTextField();
    private JTextField mRandomTF = new JTextField();
    private JTextField mStandardDevTF = new JTextField();
    private JButton mRunButton = new JButton();
    private JButton mSaveButton = new JButton();
    private JScrollPane mResultsSP = new JScrollPane();
    private JScrollPane mSettingsSP = new JScrollPane();
    private JTable mSettingsTable;
    private JPanel mCommentPanel = new JPanel();
    private JScrollPane mCommentSP = new JScrollPane();
    private JTextArea mCommentTA = new JTextArea();
    private JPanel mRestorePanel = new JPanel();
    private JButton mRestoreButton = new JButton();
    private JTextArea mResultsTA = new JTextArea();
    private JSplitPane mMainSplitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
    private JSplitPane mEastSplitPane = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
    
    // The datasets plotted in the JFreeChart ChartPanels
    private XYDataset mGeneratedClusterData, mKMeansClusterData, mXMeansClusterData;
    
    // Set to true when kmeans is running
    private boolean mRunning;
    // The two clustering objects, one for the k-means algorithm,
    // one for x-means.
    private Clustering mKMeans, mXMeans;
    // Contains the data to be clustered.
    private TestData mTestData;

    // The table model used by the JTable displaying saved settings.
    private ClusteringDemoSettingsTableModel mSettingsTableModel;

    // Provides communication with the MediaWiki server, both for downloading
    // saved settings and saving the current settings.
    private MediaWikiComm mComm;
    
    /**
     * Constructor
     * 
     * @param pageURL  the URL of the MediaWiki page containing the clustering extension, 
     * 			       expressed as a String.
     * @param extensionTag  the extension tag used with the MediaWiki server
     * @param dataTag the XML tag that surrounds the data on the MediaWiki page's edit text.
     * @param extensionType the type of the extension, "tag", "function", or "template". This is how
     *   the applet was embedded into the edit text of the page.
     * @param cookiePairs  pairs of cookies to pass to the MediaWiki server with all requests.
     * @param numCoords  the number of coordinates to initially use.
     * @param numClusters  the number of clusters to initially use.
     * @param randomSeed  the random seed to initially use.
     * @param standardDev  the standard deviation to initially use.
     */
    public ClusteringDemoPanel(
    		String pageURL, String extensionTag, String dataTag, String extensionType,
                String[] cookiePairs,
    		int numCoords, int numClusters, long randomSeed, double standardDev) {

    	// If some of the initial values make no sense, set them to their defaults.
    	if (numCoords <=  0) {
            numCoords = DEFAULT_COORDS;
        }
        if (numClusters <= 0 || numClusters > numCoords) {
            numClusters = Math.min(DEFAULT_CLUSTERS, numCoords);
        }
        if (standardDev < 0.0) {
            standardDev = DEFAULT_STANDARD_DEV;
        }
        
        mNorthPanel.setLayout(new GridBagLayout());

        mCoordLabel.setText("Number of points:");
        mClusterLabel.setText("Number of clusters:");
        mRandomLabel.setText("Random seed:");
        mStandardDevLabel.setText("Standard deviation:");

        mCoordTF.setDocument(new NumberDocument(false, false));
        mCoordTF.setText(String.valueOf(numCoords));
        mCoordTF.setColumns(10);

        mClustersTF.setDocument(new NumberDocument(false, false));
        mClustersTF.setText(String.valueOf(numClusters));
        mClustersTF.setColumns(10);

        mRandomTF.setDocument(new NumberDocument(false, false));
        mRandomTF.setText(String.valueOf(randomSeed));
        mRandomTF.setColumns(10);

        mStandardDevTF.setDocument(new NumberDocument(false, true));
        mStandardDevTF.setText(String.valueOf(standardDev));
        mStandardDevTF.setColumns(10);

        mRunButton.setText("Generate Test Data and Cluster");
        mRunButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                runClustering();
            }
        });
        
        mSaveButton.setText("Save Settings...");
        mSaveButton.addActionListener(new ActionListener() {
        	public void actionPerformed(ActionEvent e) {
        		saveSettings();
        	}
        });

        mResultsTA.setText("");
        mResultsTA.setEditable(false);
        // So the scrollpane doesn't grow vertically when the text area holds a lot of text
        // and the applet is resized.
        mResultsSP.setPreferredSize(new Dimension(1, 1));
        
        mResultsSP.setBorder(BorderFactory.createCompoundBorder(BorderFactory.createEmptyBorder(0, 10, 10, 10), 
        		BorderFactory.createTitledBorder("Clustering Messages")));
        mResultsSP.getViewport().add(mResultsTA, null);
        
        mCenterPanel.setBorder(BorderFactory.createEtchedBorder());
        
        // This places two ChartPanels in mCenterPanel.
        initPlots();

        mSettingsTableModel = new ClusteringDemoSettingsTableModel();
        mSettingsTable = new JTable(mSettingsTableModel);

        mSettingsTable.getSelectionModel().setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        mSettingsTable.setRowSelectionAllowed(true);
        mSettingsTable.getSelectionModel().addListSelectionListener(new ListSelectionListener() {
        		public void valueChanged(ListSelectionEvent e) {
        			if (!e.getValueIsAdjusting()) {
        				updateComment();
        			}
        		}
        });
        
        mSettingsSP.getViewport().add(mSettingsTable, null);
        mSettingsSP.setBorder(BorderFactory.createCompoundBorder(
        		BorderFactory.createEmptyBorder(10, 10, 5, 10), 
        		BorderFactory.createTitledBorder("Saved Settings")));
        
        mCommentSP.setBorder(BorderFactory.createCompoundBorder(
        		BorderFactory.createEmptyBorder(0, 10, 5, 10), 
        		BorderFactory.createTitledBorder("Comment")));
        mCommentSP.getViewport().add(mCommentTA, null);
        mCommentTA.setRows(3);
        mCommentTA.setEditable(false);

        mRestoreButton.setText("Restore Settings");
        mRestoreButton.addActionListener(new ActionListener() {
        	public void actionPerformed(ActionEvent e) {
        		restoreSelectedSettings();
        	}
        });
        mRestorePanel.add(mRestoreButton);
        
        mCommentPanel.setLayout(new BorderLayout());
        mCommentPanel.add(mCommentSP, BorderLayout.CENTER);
        mCommentPanel.add(mRestorePanel, BorderLayout.SOUTH);
        
        mPrevRunPanel.setLayout(new BorderLayout());
        mPrevRunPanel.add(mSettingsSP, BorderLayout.CENTER);
        mPrevRunPanel.add(mCommentPanel, BorderLayout.SOUTH);
        
        JPanel upperButtonPanel = new JPanel();
        upperButtonPanel.setLayout(new GridLayout(1, 2, 5, 0));
        upperButtonPanel.add(mRunButton);
        upperButtonPanel.add(mSaveButton);
        
        mNorthPanel.add(mCoordLabel, new GridBagConstraints(0, 0, 1, 1, 0.0,
                0.0, GridBagConstraints.EAST, GridBagConstraints.NONE,
                new Insets(10, 10, 5, 5), 0, 0));
        mNorthPanel.add(mClusterLabel, new GridBagConstraints(0, 1, 1, 1, 0.0,
                0.0, GridBagConstraints.EAST, GridBagConstraints.NONE,
                new Insets(0, 10, 5, 5), 0, 0));
        mNorthPanel.add(mRandomLabel, new GridBagConstraints(0, 2, 1, 1, 0.0,
                0.0, GridBagConstraints.EAST, GridBagConstraints.NONE,
                new Insets(0, 10, 5, 5), 0, 0));
        mNorthPanel.add(mStandardDevLabel, new GridBagConstraints(0, 3, 1, 1,
                0.0, 0.0, GridBagConstraints.EAST, GridBagConstraints.NONE,
                new Insets(0, 10, 10, 5), 0, 0));
        // The size padding is set to 200 on this field to make the split pane divider
        // move to the right.
        mNorthPanel.add(mCoordTF, new GridBagConstraints(1, 0, 1, 1, 0.25, 0.0,
                GridBagConstraints.WEST, GridBagConstraints.HORIZONTAL,
                new Insets(10, 0, 5, 10), 200, 0));
        mNorthPanel.add(mClustersTF, new GridBagConstraints(1, 1, 1, 1, 0.0,
                0.0, GridBagConstraints.WEST, GridBagConstraints.HORIZONTAL,
                new Insets(0, 0, 5, 10), 0, 0));
        mNorthPanel.add(mRandomTF, new GridBagConstraints(1, 2, 1, 1, 0.0, 0.0,
                GridBagConstraints.WEST, GridBagConstraints.HORIZONTAL,
                new Insets(0, 0, 5, 10), 0, 0));
        mNorthPanel.add(mStandardDevTF, new GridBagConstraints(1, 3, 1, 1, 0.0,
                0.0, GridBagConstraints.WEST, GridBagConstraints.HORIZONTAL,
                new Insets(0, 0, 10, 10), 0, 0));
        mNorthPanel.add(upperButtonPanel, new GridBagConstraints(0, 4, 2, 1, 1.0,
                0.0, GridBagConstraints.CENTER, GridBagConstraints.HORIZONTAL,
                new Insets(0, 10, 10, 10), 0, 0));
        
        mNorthPanel.add(mEastSplitPane, new GridBagConstraints(0, 5, 2, 1, 0.0,
                1.0, GridBagConstraints.CENTER, GridBagConstraints.BOTH,
                new Insets(0, 0, 0, 0), 0, 0));

        mEastSplitPane.setBorder(null);
        mEastSplitPane.setTopComponent(mResultsSP);
        mEastSplitPane.setBottomComponent(mPrevRunPanel);
        
        mMainSplitPane.setLeftComponent(mNorthPanel);
        mMainSplitPane.setRightComponent(mCenterPanel);
        
        this.setLayout(new BorderLayout());
        this.add(mMainSplitPane, BorderLayout.CENTER);

        if (pageURL != null && extensionTag != null && dataTag != null && extensionType != null) {
            // Instantiate the server communication object.
            mComm = new MediaWikiComm(pageURL, extensionTag, dataTag, extensionType, cookiePairs);
        }
    }
    
    /**
     * Downloads past settings from the MediaWiki server and populates the
     * settings table.
     */
    public void downloadSettings() {
        if (mComm != null) {
            try {
                mSettingsTableModel.setData(mComm.loadData());
            } catch (IOException ioe) {
                JOptionPane.showMessageDialog(this, 
                        "Error loading settings from server: " + ioe.getMessage(),
                        "Error Loading Settings", JOptionPane.ERROR_MESSAGE);
            }
        }
    }
    
    /**
     * Uploads the current settings from the settings table to the server.
     * 
     * @param retry  if true, a failure of the save caused by a concurrent edit
     *   will be followed by a reattempt after merging of the server and client settings.
     */
    public void uploadSettings(boolean retry) {

        if (mComm == null) return;
        
    	ClusteringDemoSettings[] settings = mSettingsTableModel.getData();
    	// Will be set if any exceptions are caught.
    	String errorMsg = null;

    	try {
    		
    		// Attempt the save.
    		mComm.saveData(settings);
    		
    		// The data needs to be reloaded from the server in order to get the updated
    		// edit form elements wpEditTime and wpStartTime, so futures saves succeed.
   		    try {
    			mComm.loadData();
    		} catch (IOException ioe3) {
    			String imsg = ioe3.getMessage();
    			errorMsg = "refresh of edit form data failed after successful save";
    			if (imsg != null) {
    				errorMsg += " (" + imsg + ")";
    			}
    		}
    	
    	} catch (MediaWikiComm.ConcurrentEditException cee) {
    		
    		if (retry) {

    			// There has been an edit by another user since data was downloaded from
    			// the server.  Download again, merge with the current data, and reattempt the save
    			//
    			try {

    				// Downloads the settings from the server and updates the
    				// edit form elements.
    				ClusteringDemoSettings[] freshSettings = mComm.loadData();
    	
    				// For building the new settings array.
    				java.util.List settingsList = new ArrayList();
    				// Redundant structure containing same items as settingsList.  
    				// Used to avoid inefficient linear searches of settingsList.
    				Set settingsSet = new HashSet();
    				
    				// Populate settingsList from both settings arrays, being sure to
    				// avoid duplicates.
    				for (int i=0; i<freshSettings.length; i++) {
    					ClusteringDemoSettings cds = freshSettings[i];
    					if (!settingsSet.contains(cds)) {
    						settingsList.add(cds);
    						settingsSet.add(cds);
    					}
    				}
    				for (int i=0; i<settings.length; i++) {
    					ClusteringDemoSettings cds = settings[i];
    					if (!settingsSet.contains(cds)) {
    						settingsList.add(cds);
    						settingsSet.add(cds);
    					}
    				}
    				
    				// Convert the settings list to an array and repopulate the
    				// settings table.
    				int sz = settingsList.size();
    				settings = new ClusteringDemoSettings[sz];
    				settingsList.toArray(settings);
    				mSettingsTableModel.setData(settings);

    				// Reattempt the save via a recursive call to this method, 
    				// but pass retry == false to avoid the possibility of a nasty failure loop.
    				uploadSettings(false);

    			} catch (IOException ioe2) {

    				String imsg = ioe2.getMessage();
    				errorMsg = "could not refresh data before reattempting the save";
    				if (imsg != null) {
    					errorMsg += " (" + imsg + ")";
    				}
    			}
    		}
    		
    	} catch (IOException ioe) {

    		String imsg = ioe.getMessage();
    		errorMsg = "could not save data on the server";
    		if (imsg != null) {
    			errorMsg += " (" + imsg + ")";
    		}
    	}

    	// If something went wrong, display an error message to the user.
    	if (errorMsg != null) {
    		JOptionPane.showMessageDialog(this, "Error uploading settings to the server:\n\n" + 
    				errorMsg, 
    				"Error Uploading Settings",
    				JOptionPane.ERROR_MESSAGE);
    	}
    }

    private boolean mDividerInitialized;

    /**
     * Overriden simply to set the divider location of the left splitpane after
     * realization of the panel.
     */
    public void setBounds(int x, int y, int width, int height) {
        super.setBounds(x, y, width, height);
        if (!mDividerInitialized) {
        	// One of the quirks of swing is that JSplitPane.setDividerLocation has no
        	// effect until after the split pane is realized on the screen.
            mEastSplitPane.setDividerLocation(4*height/9);
            mDividerInitialized = true;
        }
    }

    /**
     * Initializes the JFreeChart-based plots.
     */
    private void initPlots() {
        mGeneratedClusterData = new XYDataset();
        mKMeansClusterData = new XYDataset();
        mXMeansClusterData = new XYDataset();
        mCenterPanel.setLayout(new GridLayout(3, 1));
        mCenterPanel.add(initChartPanel(mGeneratedClusterData, "Generated Clusters"));
        mCenterPanel.add(initChartPanel(mKMeansClusterData, "K-Means Clusters"));
        mCenterPanel.add(initChartPanel(mXMeansClusterData, "X-Means Clusters"));
    }
    
    /**
     * Creates a ChartPanel from the JFreeChart library.
     * 
     * @param dataset  the dataset to plot.
     * @param title  the plot's title.
     * 
     * @return  a ChartPanel object.
     */
    private ChartPanel initChartPanel(XYDataset dataset, String title) {
        
        NumberAxis xAxis = new NumberAxis("X");
        NumberAxis yAxis = new NumberAxis("Y");
        
        xAxis.setAutoRangeIncludesZero(false);
        yAxis.setAutoRangeIncludesZero(false);
        xAxis.setAutoRange(true);
        yAxis.setAutoRange(true);
        
        XYItemRenderer renderer = new XYLineAndShapeRenderer(false, true);
        
        renderer.setShape(new java.awt.geom.Rectangle2D.Double(-1.0, -1.0, 2.0, 2.0));
        
        XYPlot plot = new XYPlot(dataset, xAxis, yAxis, renderer);
        
        plot.setOrientation(PlotOrientation.VERTICAL);

        JFreeChart chart = new JFreeChart(title, JFreeChart.DEFAULT_TITLE_FONT,
                plot, false);
        
        return new ChartPanel(chart);
    }
    
    /**
     * Launch the clustering algorithms.
     */
    private void runClustering() {
    
        if (!mRunning) {
            
            try {
                
            	// Parse the current settings.
                int numCoords = getEnteredValue(mCoordTF, 1, 3000);
                int numClusters = getEnteredValue(mClustersTF, 1, (numCoords - 1));
                long seed = getEnteredValue(mRandomTF, Long.MIN_VALUE, Long.MAX_VALUE);
                double standardDev = getEnteredValue(mStandardDevTF, 0.0, 1.0);
            
                // Generate the test data.
                mTestData = new TestData(numCoords, 2, numClusters, standardDev, seed);
                
                // Clear the plots.
                mGeneratedClusterData.clear();
                mKMeansClusterData.clear();
                mXMeansClusterData.clear();
                
                // Populate the generated data plot.
                double[][] coordinates = mTestData.getCoordinates();
                Cluster[] clusters = mTestData.getClusters();
                Arrays.sort(clusters, new CenterComparator());
                
                for (int i=0; i<numClusters; i++) {
                    mGeneratedClusterData.addData("C" + (i+1),
                            coordinates, clusters[i].getMemberIndexes());
                }
        
                // Clear the results text area.
                mResultsTA.setText("");
                // Disable the run button.
                mRunButton.setEnabled(false);
                mRunning = true;
                
                // Launch k-means.  
                launchKMeans(numClusters, seed);
                
            } catch (InsufficientMemoryException e) {
                
                displayInsufficientMemoryDialog(e.getMessage());
                
            } catch (Exception e) {
            
                e.printStackTrace();
                
                JOptionPane.showMessageDialog(this, 
                		"There was a problem with one of the values you entered:\n\n" +
                        e.getMessage() + "\n\n" +
                        "Please correct the problem and try again.", "Error", JOptionPane.ERROR_MESSAGE);
                
                mRunning = false;
            }
        }

    }
    
    private void saveSettings() {
    	
        try {
            
            int numCoords = getEnteredValue(mCoordTF, 1, 3000);
            int numClusters = getEnteredValue(mClustersTF, 1, (numCoords - 1));
            long seed = getEnteredValue(mRandomTF, Long.MIN_VALUE, Long.MAX_VALUE);
            double standardDev = getEnteredValue(mStandardDevTF, 0.0, 1.0);
        
            ClusteringDemoSettings oldSettings = mSettingsTableModel.findSettings(
            		numCoords, numClusters, standardDev, seed);
            if (oldSettings != null) {
            	if (JOptionPane.showConfirmDialog(this, "The same settings were saved on " +
            			mSettingsTableModel.formatTimeStamp(oldSettings.getTimeStamp()) + ".\nSave anyway?",
            			"Save Settings", JOptionPane.YES_NO_OPTION) == JOptionPane.NO_OPTION) {
            		return;
            	}
            }
            
            String comment = JOptionPane.showInputDialog("Please enter a comment: ");
            if (comment != null) { 
            	mSettingsTableModel.addSettings(new ClusteringDemoSettings(numCoords,
            			numClusters, standardDev, seed, System.currentTimeMillis(),
            			comment
            			));
            	uploadSettings(true);
            }
            
        } catch (Exception e) {
        
        	e.printStackTrace();
        	
            JOptionPane.showMessageDialog(this, "There was a problem with one of the values you entered:\n\n" +
                    e.getMessage() + "\n\n" +
                    "Please correct the problem and try again.", "Error", JOptionPane.ERROR_MESSAGE);
            
            mRunning = false;
        }
    }
    
    private void updateComment() {
    	int row = mSettingsTable.getSelectedRow();
    	String comment = row >= 0 ? mSettingsTableModel.getComment(row) : "";
    	mCommentTA.setText(comment);
    }
    
    private void restoreSelectedSettings() {
    	int row = mSettingsTable.getSelectedRow();
    	if (row >= 0) {
    		ClusteringDemoSettings settings = mSettingsTableModel.getSettings(row);
    		mCoordTF.setText(String.valueOf(settings.getCoordinates()));
    		mClustersTF.setText(String.valueOf(settings.getClusters()));
    		mRandomTF.setText(String.valueOf(settings.getRandomSeed()));
    		mStandardDevTF.setText(String.valueOf(settings.getStandardDev()));
    	}
    }
    
    private void launchKMeans(int numClusters, long seed) {
        
        mKMeans = new BasicKMeans(mTestData.getCoordinates(), numClusters, 500, seed);
        mKMeans.addClusteringListener(this);

        Thread t = new Thread(mKMeans);
        t.setDaemon(true);
        t.start();        
    }
    
    private void launchXMeans(long seed) {        
        mXMeans = new XMeans(mTestData.getCoordinates(), seed);
        mXMeans.addClusteringListener(this);
        
        Thread t = new Thread(mXMeans);
        t.setDaemon(true);
        t.start();
    }
    
    /**
     * Displays an error dialog stating that insufficient memory is
     * available.
     */
    private void displayInsufficientMemoryDialog(String message) {
        
        String dlgMessage = null;
        if (message != null && message.length() > 0) {
            dlgMessage = "Insufficient memory is available (" + message + ").\n" +
                "Try reducing the number of coordinates and/or the number of clusters.";
        } else {
            dlgMessage = 
                "Insufficient memory is available.  Try reducing the \n" + 
                "number of coordinates and/or the number of clusters.";
        }
        
        JOptionPane.showMessageDialog(this, 
                dlgMessage,
                "Insufficient Memory", JOptionPane.ERROR_MESSAGE);
    }

    private void displayText(String text) {
        mResultsTA.append(text);
        if (!text.endsWith("\n")) {
            mResultsTA.append("\n");
        }
    }

    /**
     * Cleanup after completion of k-means.
     */
    private void cleanupAfterKMeans(boolean launchXMeans) {
        long seed = ((BasicKMeans) mKMeans).getRandomSeed();
        if (mKMeans != null) {
            mKMeans.removeClusteringListener(this);
            mKMeans = null;
        }
        if (launchXMeans) {
            launchXMeans(seed);
        } else {
            mTestData = null;
            mRunning = false;
            mRunButton.setEnabled(true);
        }
    }
    
    /**
     * Cleanup after completion of x-means.
     */
    private void cleanupAfterXMeans() {
        if (mXMeans != null) {
            mXMeans.removeClusteringListener(this);
            mXMeans = null;
        }
        mTestData = null;
        mRunning = false;
        mRunButton.setEnabled(true);
    }
    

    public void clusteringComplete(final Clustering src, 
            final Cluster[] clusters, final long executionTime) {
        
        if (src == mKMeans || src == mXMeans) {
            
            SwingUtilities.invokeLater(new Runnable() {

                public void run() {

                    XYDataset dataset = null;
                    
                    String prefix = null;
                    if (src == mKMeans) {
                        prefix = "K-Means complete: ";
                        dataset = mKMeansClusterData;
                    } else if (src == mXMeans) {
                        prefix = "X-Means complete: ";
                        dataset = mXMeansClusterData;
                    }

                    displayText(prefix + "processing time (ms) = " + executionTime);
                    displayText("Number of clusters: " + clusters.length);
                    if (src == mKMeans) {
                        displayText("\n");
                    }

                    // Populate the generated data plot.
                    double[][] coordinates = mTestData.getCoordinates();
                    Arrays.sort(clusters, new CenterComparator());

                    for (int i=0; i<clusters.length; i++) {
                        dataset.addData("C" + (i+1),
                                coordinates, clusters[i].getMemberIndexes());
                    }

                    if (src == mKMeans) {
                        cleanupAfterKMeans(true);
                    } else if (src == mXMeans) {
                        cleanupAfterXMeans();
                    }
                }
            });
            
        }
    }

    public void clusteringError(final Clustering src, final Throwable t) {
        if (src == mKMeans || src == mXMeans) {
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                if (t instanceof InsufficientMemoryException) {
                    displayText("K-Means aborted because of insufficient memory.");
                    displayInsufficientMemoryDialog(t.getMessage());
                } else {
                    StringWriter sw = new StringWriter();
                    PrintWriter pw = new PrintWriter(sw);
                    t.printStackTrace(pw);
                    displayText(sw.toString());
                }
                if (src == mKMeans) {
                    cleanupAfterKMeans(false);
                } else if (src == mXMeans) {
                    cleanupAfterXMeans();
                }
            }
        });
        }
    }

    public void clusteringMessage(final Clustering src, final String message) {
        if (src == mKMeans || src == mXMeans) {
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                displayText(message);
            }
        });
        }
    }

    /**
     * Method for validating entries typed into text fields.
     */
    private static int getEnteredValue(JTextField tf, int min, int max) {
        int value = 0;
        String s = tf.getText().trim();
        if (s.length() == 0) {
            throw new RuntimeException("blank entry");
        }
        try {
            value = Integer.parseInt(s);
            if (value < min || value > max) {
                throw new RuntimeException("outside range [" + min + " - " + max + "]: " + value);
            }
        } catch (NumberFormatException nfe) {
            throw new RuntimeException("invalid number: " + s);
        }
        return value;
    }

    /**
     * Method for validating entries typed into text fields.
     */
    private static long getEnteredValue(JTextField tf, long min, long max) {
        long value = 0L;
        String s = tf.getText().trim();
        if (s.length() == 0) {
            throw new RuntimeException("blank entry");
        }
        try {
            value = Long.parseLong(s);
            if (value < min || value > max) {
                throw new RuntimeException("outside range [" + min + " - " + max + "]: " + value);
            }
        } catch (NumberFormatException nfe) {
            throw new RuntimeException("invalid number: " + s);
        }
        return value;
    }

    /**
     * Method for validating entries typed into text fields.
     */
    private static double getEnteredValue(JTextField tf, double min, double max) {
        double value = 0;
        String s = tf.getText().trim();
        if (s.length() == 0) {
            throw new RuntimeException("blank entry");
        }
        try {
            value = Double.parseDouble(s);
            if (value < min || value > max) {
                throw new RuntimeException("outside range [" + min + " - " + max + "]: " + value);
            }
        } catch (NumberFormatException nfe) {
            throw new RuntimeException("invalid number: " + s);
        }
        return value;
    }

    private static class TestData {
        
        private double[][] mCoordinates;
        private Cluster[] mClusters;
        
        private TestData(int coordCount, int dimensions, int clusterCount, 
                double standardDev, long randomSeed) 
        throws InsufficientMemoryException {

            // Explicit garbage collection to reduce the likelihood of 
            // having insufficient memory.
            System.gc();
            
            long memRequired = 8L * (long) dimensions * (long) (coordCount + clusterCount);
            long memFree = Runtime.getRuntime().freeMemory();
            
            if (memFree < memRequired) {
                throw new InsufficientMemoryException("memory required = " + memRequired + ", free = " + memFree);
            }

            int nEachCluster = 10;
            if (clusterCount > 0) {
                nEachCluster = Math.max(10, (int) (0.9 * coordCount/clusterCount));
            }
            
            double[][] coordinates = new double[coordCount][dimensions];
            double[][] exemplars = new double[clusterCount][dimensions];
            
            double[][] centers = new double[clusterCount][dimensions];
            int[][] memberships = new int[clusterCount][nEachCluster];
            int[] counts = new int[clusterCount];

            Random random = new Random(randomSeed);
            for (int i=0; i<clusterCount; i++) {
                for (int j=0; j<dimensions; j++) {
                    exemplars[i][j] = random.nextDouble();
                }
            }
     
            for (int i=0; i<coordCount; i++) {
                
                int cluster = random.nextInt(clusterCount);
                int memIndex = counts[cluster];
                
                // Ran out of room for the membership list, so expand it a little.
                if (memberships[cluster].length == memIndex) {
                    int oldLength = memberships[cluster].length;
                    int newLength = Math.max(oldLength + 1, (int) (1.2 * oldLength));
                    int[] newMembership = new int[newLength];
                    System.arraycopy(memberships[cluster], 0, newMembership, 0, oldLength);
                    memberships[cluster] = newMembership;
                }
                
                // Record the index as a member of the cluster.
                memberships[cluster][memIndex] = i;
                // Increment the count for the cluster.
                counts[cluster]++;
                
                double[] exemplar = exemplars[cluster];
                double[] coord = coordinates[i];
                double[] center = centers[cluster];
                
                for (int j=0; j<dimensions; j++) {
                    double d = exemplar[j] + standardDev*random.nextGaussian();
                    coord[j] = d; 
                    center[j] += d;
                }
                
            }
            
            Cluster[] clusters = new Cluster[clusterCount];
            for (int i=0; i<clusterCount; i++) {
                int count = counts[i];
                if (count > 1) {
                    double[] center = centers[i];
                    for (int j=0; j<dimensions; j++) {
                        center[j] /= count;
                    }
                }
                
                int[] members = new int[count];
                System.arraycopy(memberships[i], 0, members, 0, count);
                // No longer needed.
                memberships[i] = null;
                
                clusters[i] = new Cluster(members, centers[i]);
            }
            
            mCoordinates = coordinates;
            mClusters = clusters;
        }
        
        private double[][] getCoordinates() {
            return mCoordinates;
        }
        
        private Cluster[] getClusters() {
            return mClusters;
        }
    }
    
    // A simple comparator for sorting clusters by their centers.
    private static class CenterComparator implements Comparator {

        public int compare(Object o1, Object o2) {
            Cluster c1 = (Cluster) o1;
            Cluster c2 = (Cluster) o2;
            double[] center1 = c1.getCenter();
            double[] center2 = c2.getCenter();
            int n = center1.length;
            for (int i=0; i<n; i++) {
                double d1 = center1[i];
                double d2 = center2[i];
                if (d1 < d2) return -1;
                if (d1 > d2) return +1;
            }
            return 0;
        }
        
    }
}
