package clustering;

import java.io.*;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.util.*;
import java.util.regex.*;

import javax.xml.parsers.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;

/**
 * <code>ClusteringDemoSettingsFactory</code> is a class containing
 * static utility methods for parsing and saving instances of
 * <code>ClusteringDemoSetting</code>s.
 * 
 * @author R. Scarberry
 *
 */
public class ClusteringDemoSettingsFactory {

    // Used as XML attributes.
    public static final String SETTINGS = "settings";
    public static final String COORDINATES = "coords";
    public static final String CLUSTERS = "clusters";
    public static final String STANDARD_DEV = "sdev";
    public static final String RANDOM_SEED = "seed";
    public static final String TIMESTAMP = "time";
    public static final String COMMENT = "comment";

    // The three extension types expressed as integers.
    public static final int EXTENSION_TYPE_TAG      = 0;
    public static final int EXTENSION_TYPE_FUNCTION = 1;
    public static final int EXTENSION_TYPE_TEMPLATE = 2;

    // Strings used several times in generating the XML.
    private static final String ATT_PREFIX = "=\"";
    private static final String ATT_SUFFIX = "\" ";

    /**
     * Parses an array of <code>ClusteringDemoSetting</code>s objects from 
     * a block of XML if such is contained in the first argument.
     * 
     * @param pageContent unrendered content from the wiki page gotten 
     *   from the wpTextbox1 item of the form editform. 
     *   
     * @param dataTag - the xml tag used with the extension handler, 
     *   without the angle brackets surrounding it.
     * 
     * @return - an array of settings objects, parsed from the xml. If no
     *   XML is found in pageContent, a zero-length array is returned.
     */
    public static ClusteringDemoSettings[] parseFromContent(
            String pageContent, String dataTag) {

        // Mediawiki extension tags are case insensitive, 
        // so do a case-insensitive search.  Make a lower-case copy of 
        // the page content only for searching.
        String lccontent = pageContent.toLowerCase();

        String startTag = "<" + dataTag.toLowerCase() + ">";

        // Look for the start tag.
        int start = lccontent.indexOf(startTag);

        if (start >= 0) { // Found the start tag "<clustering_data>"

            // Look for the end tag.
            String endTag = "</" + dataTag.toLowerCase() + ">";
            int end = lccontent.indexOf(endTag, start + startTag.length());
            
            if (end > start) {

                // If it was found, parse it using SAX.
                //
                SAXParserFactory factory = SAXParserFactory.newInstance();
                try {

                    SAXParser parser = factory.newSAXParser();
                    XMLReader reader = parser.getXMLReader();
                    // Class that contains the customized parsing.
                    SettingsXMLHandler handler = new SettingsXMLHandler();
                    reader.setContentHandler(handler);

                    // Extract the xml from the unmodified page content.
                    String xmlBlock = pageContent.substring(start, end + endTag.length());

                    // Kick off parsing.
                    reader.parse(new InputSource(new StringReader(xmlBlock)));
                    // Return the objects from the instance of SettingsXMLHandler.
                    return handler.getSettings();

                } catch (Exception e) {
                    // Don't let the error be fatal, but show the stack trace in the console.
                    e.printStackTrace();
                }

            }
        }

        // Either the parse was aborted by an exception, or there was no data
        // block.  Return an empty array.
        return new ClusteringDemoSettings[0];
    }

    /**
     * Returns a new version of its first argument containing an updated
     * XML block for the settings objects.
     * 
     * @param pageContent   the old edit tab contents.
     * @param extensionTag  the tag used for embedding the applet extension.
     * @param dataTag       the tag used for the data XML block.
     * @param extensionType the manner in which the applet extension was embedded.
     * @param settings      the array of settings objects to be transformed into
     *   XML and saved.
     *   
     * @return an updated version of pageContent with a new XML block for
     *   the settings objects.
     */
    public static String updateContent(String pageContent, 
            String extensionTag, String dataTag, int extensionType, 
            ClusteringDemoSettings[] settings) {

        // Simply look for an old XML block for the settings objects.
        // If found, this method simply replaces the block with a new one.
        //
        // If not found, then this method has to figure out where to insert
        // the block, which is much more complicated.
        
        
        // Mediawiki extension tags are case insensitive, 
        // so do a case-insensitive search.
        String lccontent = pageContent.toLowerCase();

        // Set startTag to something of the form "<clustering_data>"
        String startTag = "<" + dataTag.toLowerCase() + ">";

        int start = lccontent.indexOf(startTag);
        if (start >= 0) { // Found it.
            // Look for the closing tag.
            String endTag = "</" + dataTag.toLowerCase() + ">";
            int end = lccontent.indexOf(endTag, start + startTag.length());
            if (end > start) { // Found the closing tag.
                // A data block was already in the page content.  Simply replace
                // the old block with a new one.
                StringBuffer sb = new StringBuffer(pageContent.length());
                // Append the text that prefixes the old data block.
                sb.append(pageContent.substring(0, start));
                // Append the new data block.
                sb.append(toXML(dataTag, settings));
                // Append the text that follows the old data block.
                sb.append(pageContent.substring(end + endTag.length()));
                
                // Return the new page content.
                return sb.toString();
            }
        }

        // If execution has fallen through, there was no data block on the page,
        // so we have to figure out where to put it.  We'll put it immediately under
        // the expression that embeds the applet.  
        // (extension tag, parser function, or template).

        // Find the place using regular expression matching.
        String[] regexps = regularExpressions(extensionTag, extensionType);
        // Cycle through the regular expressions until a match is found.
        for (int i=0; i<regexps.length; i++) {
            
            try {
                
                // Need to use the DOTALL flag so the "." char matches newlines.
                Pattern pattern = Pattern.compile(regexps[i], Pattern.CASE_INSENSITIVE | Pattern.DOTALL);
                Matcher m = pattern.matcher(pageContent);
                
                if (m.matches()) { // Found a match
                    
                    // The insertion point is just after the phrase matching 
                    // group 1.  This phrase is "<clustering>...</clustering>" or
                    // "{{#clustering:2000}}", "{{clusteringTemplate}}", etc., depending on
                    // the extensionType.
                    int insertionPoint = m.end(1);
                    
                    StringBuffer sb = new StringBuffer(pageContent.length());
                    // Append everything before the insertion point. This will
                    // include the phrase.
                    sb.append(pageContent.substring(0, insertionPoint));
                    // Append the data xml.
                    sb.append(toXML(dataTag, settings));
                    // Append everything that came after the insertion point.
                    sb.append(pageContent.substring(insertionPoint));
                    
                    // Done.
                    return sb.toString();
                    
                }
                
            } catch (PatternSyntaxException pse) {
                // Developer tested -- doesn't happen.
            }
        }
        
        // Nothing worked, so return null.
        return null;
    }

    /**
     * Returns regular expressions for matching the text used to embed the applet extension
     * in the webpage.  The regular expressions need to place the match in group 1.
     * 
     * @param extensionTag
     * @param extensionType
     * 
     * @return
     */
    private static String[] regularExpressions(String extensionTag, int extensionType) {
        switch(extensionType) {
        case EXTENSION_TYPE_TAG:
            // This case returns two regular expression, one for
            // <clustering>...</clustering> and one for 
            // <clustering ... />
            return new String[] {
                    ".*?(<" + extensionTag + "\\b([^>]*?)>(.*?)</" + extensionTag + ">).*",
                    ".*?(<" + extensionTag + "\\b([^>]*?)/>).*"                    
            };
        case EXTENSION_TYPE_FUNCTION:
            return new String[] {
                    ".*?(\\{\\{\\s*#" + extensionTag + "\\s*:([^}]*?)\\}\\}).*",
            };
        case EXTENSION_TYPE_TEMPLATE:
            return new String[] {
                    ".*?(\\{\\{\\s*" + extensionTag + "\\s*([^:}]*?)\\}\\}).*"
            };
        }
        // Shouldn't ever fall through.
        return new String[0];
    }
    
    /**
     * Generates an XML version of the array of settings objects.
     * 
     * @param dataTag  tag used to open and close the XML block.
     * @param settings array of settings objects.
     * @return
     */
    private static String toXML(String dataTag, ClusteringDemoSettings[] settings) {
        StringBuffer sb = new StringBuffer();
        sb.append("<" + dataTag + ">");
        int n = settings.length;
        for (int i=0; i<n; i++) {
            sb.append(toXML(settings[i]));
        }
        sb.append("</" + dataTag + ">");
        return sb.toString();
    }
    
    /**
     * Generates XML for one ClusteringDemoSettings object.
     * @param settings 
     * @return
     */
    private static String toXML(ClusteringDemoSettings settings) {

        StringBuffer sb = new StringBuffer();

        sb.append('<');
        sb.append(SETTINGS);
        sb.append(' ');
        sb.append(COORDINATES);
        sb.append(ATT_PREFIX);
        sb.append(settings.getCoordinates());
        sb.append(ATT_SUFFIX);
        sb.append(CLUSTERS);
        sb.append(ATT_PREFIX);
        sb.append(settings.getClusters());
        sb.append(ATT_SUFFIX);
        sb.append(STANDARD_DEV);
        sb.append(ATT_PREFIX);
        sb.append(settings.getStandardDev());
        sb.append(ATT_SUFFIX);
        sb.append(RANDOM_SEED);
        sb.append(ATT_PREFIX);
        sb.append(settings.getRandomSeed());
        sb.append(ATT_SUFFIX);
        sb.append(TIMESTAMP);
        sb.append(ATT_PREFIX);
        sb.append(settings.getTimeStamp());
        sb.append(ATT_SUFFIX);
        sb.append(COMMENT);

        String comment = settings.getComment();
        if (comment == null) {
            comment = "";
        }

        sb.append(ATT_PREFIX);
        try {
            sb.append(URLEncoder.encode(comment, "UTF-8"));
        } catch (UnsupportedEncodingException uee) {
            // Unlikely to not recognize "UTF-8"!
        }

        sb.append(ATT_SUFFIX);
        sb.append("/>");

        return sb.toString();
    }

    /**
     * Class implementing customized SAX parsing of a block of 
     * XML generated from <code>ClusteringDemoSetting</code>s objects.
     * 
     * @author R. Scarberry
     *
     */
    private static class SettingsXMLHandler extends DefaultHandler {

        // List in which to accumulate the settings objects.
        private List mSettingsList = new ArrayList();

        /**
         * Returns the parsed <code>ClusteringDemoSetting</code>s objects.
         * 
         * @return an array, never null.
         */
        private ClusteringDemoSettings[] getSettings() {
            int sz = mSettingsList.size();
            ClusteringDemoSettings[] settings = new ClusteringDemoSettings[sz];
            if (sz > 0) {
                mSettingsList.toArray(settings);
            }
            return settings;
        }

        /**
         * Since all data items are contained in the element attributes,
         * startElement is the only method necessary.  There is no need for
         * endElement() or characters().
         */
        public void startElement(String uri, String localName, String qName,
                Attributes attributes) throws SAXException {

            // The tag, converted to lowercase.
            String tag = localName.length() > 0 ? localName.toLowerCase() : qName.toLowerCase();

            // Only pay attention to "<settings>"
            if (tag.equals(SETTINGS)) {

                int coordinates = -1, clusters = -1;
                double standardDev = -1;
                long randomSeed = -1L, timeStamp = -1L;
                String comment = null;

                int n = attributes.getLength();

                // The values of all ClusteringDemoSettings object variables are
                // passed as attributes to the settings tag.
                //
                for (int i=0; i<n; i++) {
                    String aname = attributes.getLocalName(i);
                    if (aname.length() == 0) {
                        aname = attributes.getQName(i);
                    }
                    aname = aname.toLowerCase();
                    String value = attributes.getValue(i).trim();
                    try {
                        if (COORDINATES.equals(aname)) {
                            coordinates = Integer.parseInt(value);
                        } else if (CLUSTERS.equals(aname)) {
                            clusters = Integer.parseInt(value);
                        } else if (STANDARD_DEV.equals(aname)) {
                            standardDev = Double.parseDouble(value);
                        } else if (RANDOM_SEED.equals(aname)) {
                            randomSeed = Long.parseLong(value);
                        } else if (TIMESTAMP.equals(aname)) {
                            timeStamp = Long.parseLong(value);
                        } else if (COMMENT.equals(aname)) {
                            // Since the comment is entered by the user, it should
                            // be encoded when stored and decoded when parsed.
                            comment = URLDecoder.decode(value, "UTF-8");
                        }
                    } catch (Exception e) { // NumberFormatExceptions, etc..
                        e.printStackTrace();
                    }
                }

                // If everything has been parsed successfully, create a new settings
                // object and add it to the list.
                if (coordinates > 0 && clusters > 0 && standardDev >= 0 && 
                        randomSeed >= 0L && timeStamp >= 0L) {
                    if (comment == null) {
                        comment = "";
                    }
                    mSettingsList.add(
                            new ClusteringDemoSettings(coordinates, clusters, standardDev, 
                                    randomSeed, timeStamp, comment));
                }
            }
        }
    }
}
