package clustering;

import java.awt.BorderLayout;
import java.util.ArrayList;

import javax.swing.JApplet;
import javax.swing.SwingUtilities;

/**
 * KMeansApplet to demo parallelization of KMeans clustering.
 * 
 * @author R. Scarberry
 *
 */
public class DemoApplet extends JApplet {

    public static final String TITLE_SECTION = "?title=";

    public void init() {
        
        // Retrieve 3 args to the KMeansPanel constructor from the
        // applet parameters, supplying sensible defaults.
        int coordinates = getIntParameter("coordinates", 1000);
        int clusters = getIntParameter("clusters", 10);
        double standardDev = getDoubleParameter("standardDev", 0.03);
        long randomSeed = getLongParameter("randomSeed", 1234L);
        
        String extensionTag = getParameter("extensionTag", null);
        String dataTag = getParameter("dataTag", null);
        String extensionType = getParameter("extensionType", "tag");
        
        String[] cookiePairs = parseCookiePairs();
        
        // This could be either of two forms:
        //
        // a) http://hostname/mediawiki/index.php/pagetitle
        //    or
        // b) http://hostname/mediawiki/index.php?title=pagetitle
        //
        String docBase = getDocumentBase().toString();
        String pageURL = docBase;
        
        int n = docBase.lastIndexOf('/');

        if (n > 0) { // Should always be the case...

        	// Case a):  base = http://hostname/mediawiki/index.php
        	//           pageTitle = pagetitle
        	//
        	// Case b):  base = http://hostname/mediawiki
        	//           pageTitle = index.php?title=pagetitle
        	//
            String base = docBase.substring(0, n);
            String pageTitle = docBase.substring(n+1);

            // After this block, for either case a) or case b):
            //
            // base = http://hostname/mediawiki/index.php?title=
            // pageTitle = pagetitle
            //
            n = pageTitle.indexOf(TITLE_SECTION);
            if (n >= 0) { // It is...
                base += "/" + pageTitle.substring(0, n + TITLE_SECTION.length());
                pageTitle = pageTitle.substring(n + TITLE_SECTION.length());
            } else {
                base += TITLE_SECTION;
            }

            // String off any extra anchors or parameters from pageTitle.
            //
            n  = pageTitle.indexOf("#"); // strip off subpage name if it exists...
            if (n > -1 ) {
                pageTitle = pageTitle.substring(0, n);
            }
            n = pageTitle.indexOf('&');
            if (n > -1) {
                pageTitle = pageTitle.substring(0, n);
            }
            
            // Add for the complete url specification.
            //
            pageURL = base + pageTitle;
        }
        
        final ClusteringDemoPanel demoPanel = new ClusteringDemoPanel(
        	pageURL, extensionTag, dataTag, extensionType, cookiePairs,
        	coordinates, clusters, randomSeed, standardDev
        );
        
        // Set the KMeansPanel as the center component.
        this.getContentPane().add(demoPanel, BorderLayout.CENTER);
        
        SwingUtilities.invokeLater(new Runnable() {
        	public void run() {
        		demoPanel.downloadSettings();
        	}
        });
    }
    
    private String getParameter(String name, String def) {
        String returnValue = getParameter(name);
        if (returnValue == null) {
        	returnValue = def;
        }
        return returnValue;     
    }

    private int getIntParameter(String name, int def) {
        int returnValue = def;
        String param = getParameter(name);
        if (param != null) {
            try {
                returnValue = Integer.parseInt(param);
            } catch (NumberFormatException nfe) {
                // Return the default
            }
        }
        return returnValue;     
    }

    private long getLongParameter(String name, long def) {
        long returnValue = def;
        String param = getParameter(name);
        if (param != null) {
            try {
                returnValue = Long.parseLong(param);
            } catch (NumberFormatException nfe) {
                // Return the default
            }
        }
        return returnValue;     
    }
    
    private double getDoubleParameter(String name, double def) {
        double returnValue = def;
        String param = getParameter(name);
        if (param != null) {
            try {
                returnValue = Double.parseDouble(param);
            } catch (NumberFormatException nfe) {
                // Return the default
            }
        }
        return returnValue;     
    }
    
    /**
     * Parses cookies from the applet parameters that are relevant to
     * applet-server communication.  The server-side PHP script passes the
     * cookies to the applet as parameters.  The ones that are used with
     * client-server communication are <pre>UserID, UserName, and _session</pre>.
     * 
     * @return - an array of cookie name/value pairs.  The ones with even indexes
     *   contain the names, the ones with odd indexes contain the values.
     */
    private String[] parseCookiePairs() {

    	String empty = "";

    	// The prefix used for mediawiki cookies is passed as a separate parameter.
        String prefix = getParameter("cookiePrefix", empty);
        String cookieName = null;
        
        String param = empty;
        java.util.List cookieList = new ArrayList();

        cookieName = "UserID";
        param = getParameter(cookieName, empty);
        
        if (param.length() > 0) {
        	cookieList.add(prefix + cookieName);
        	cookieList.add(param);
        }

        cookieName = "UserName";
        param = getParameter(cookieName, empty);
        
        if (param.length() > 0) {
        	cookieList.add(prefix + cookieName);
        	cookieList.add(param);
        }

        cookieName = "_session";
        param = getParameter(cookieName, empty);
        
        if (param.length() > 0) {
        	cookieList.add(prefix + cookieName);
        	cookieList.add(param);
        }

        int sz = cookieList.size();
        String[] cookiePairs = new String[sz];
        if (sz > 0) {
        	cookieList.toArray(cookiePairs);
        }

        return cookiePairs;
    }
}
