package clustering;

import java.util.*;

public class XMeans implements Clustering {

    public static final double LOG2PI = Math.log(2 * Math.PI);

    private double[][] mCoordinates;

    private Cluster[] mClusters;

    private long mRandomSeed;

    private Set mUnsplittables = new HashSet();

    // Listeners to be notified of significant happenings.
    private List mListeners = new ArrayList(1);

    public XMeans(double[][] coordinates, long randomSeed) {
        mCoordinates = coordinates;
        mRandomSeed = randomSeed;
    }

    private static Cluster[] prepareClusterArray(Cluster[] clusters, int omit,
            Cluster[] splitClusters) {
        int numClusters = clusters.length;
        int sz = numClusters + splitClusters.length - 1;
        Cluster[] rtn = new Cluster[sz];
        int count = 0;
        for (int i = 0; i < numClusters; i++) {
            if (i != omit) {
                rtn[count++] = clusters[i];
            }
        }
        for (int i = 0; i < splitClusters.length; i++) {
            rtn[count++] = splitClusters[i];
        }
        return rtn;
    }

    private Cluster[] split(Cluster cluster, int howMany) {

        Map indexMap = null;

        double[][] coords = mCoordinates;
        int dim = mCoordinates[0].length;

        int clusterSize = cluster.getSize();

        if (clusterSize < mCoordinates.length) {
            int[] indexes = cluster.getMemberIndexes();
            // Used to map sequential indexes (0, 1, 2, ...) to the 
            // original indexes from getMemberIndexes();
            indexMap = new HashMap(2 * clusterSize);
            // Need to use a new coordinate buffer
            coords = new double[clusterSize][dim];
            for (int i = 0; i < clusterSize; i++) {
                int index = indexes[i];
                indexMap.put(new Integer(i), new Integer(index));
                System.arraycopy(mCoordinates[index], 0, coords[i], 0, dim);
            }
        }

        Clustering kmeans = new BasicKMeans(coords, howMany, Integer.MAX_VALUE,
                mRandomSeed);
        kmeans.run();

        kmeans.run();

        Cluster[] splitClusters = kmeans.getClusters();
        if (indexMap != null) {
            // Need to map the ids back to the ones from the original cluster
            for (int i = 0; i < splitClusters.length; i++) {
                int[] members = splitClusters[i].getMemberIndexes();
                for (int j = 0; j < members.length; j++) {
                    Integer orig = (Integer) indexMap.get(new Integer(
                            members[j]));
                    members[j] = orig.intValue();
                }
                splitClusters[i] = new Cluster(members, splitClusters[i]
                        .getCenter());
            }
        }

        return splitClusters;
    }

    private int splitClusters(int[] splitsToTry) {

        int numClusters = mClusters.length;

        int splits = 0;
        List newClusterList = new ArrayList();

        double startingBIC = computeBIC(mCoordinates, mClusters);

        double bicThreshold = startingBIC / 2.0;

        for (int i = 0; i < numClusters; i++) {

            Cluster cluster = mClusters[i];
            boolean keepOriginal = true;

            if (!mUnsplittables.contains(cluster)) {

                int clusterSize = cluster.getSize();
                double bicCluster = computeBIC(mCoordinates,
                        new Cluster[] { cluster });
                Cluster[] splitClusters = null;

                for (int j = 0; j < splitsToTry.length; j++) {

                    int splitDiv = splitsToTry[j];

                    if (clusterSize < splitDiv)
                        continue;

                    splitClusters = split(cluster, splitDiv);

                    double bic2 = computeBIC(mCoordinates, prepareClusterArray(
                            mClusters, i, splitClusters));

                    if (!Double.isNaN(bic2) && bic2 > startingBIC) {
                        keepOriginal = false;
                        break;
                    }
                } // for (int j...

                if (!keepOriginal) {
                    splits++;
                    for (int c = 0; c < splitClusters.length; c++) {
                        newClusterList.add(splitClusters[c]);
                    }
                } else {
                    newClusterList.add(cluster);
                    mUnsplittables.add(cluster);
                }

            } else { // clusterSize == 1 or we've tried to split it before w/o improvement.

                newClusterList.add(cluster);

            }

        }

        if (splits > 0) {
            int sz = newClusterList.size();
            mClusters = new Cluster[sz];
            newClusterList.toArray(mClusters);
        }

        return splits;
    }

    private void ensureCoordinatesInNearestCluster() {

        int numClusters = mClusters.length;
        int dim = mCoordinates.length > 0 ? mCoordinates[0].length : 0;

        // Maps clusters to HashSets containing members to remove.
        Map removalMap = new HashMap();
        // Maps clusters to HashSets containing members to add.
        Map additionMap = new HashMap();

        for (int i = 0; i < numClusters; i++) {

            Cluster cluster = mClusters[i];
            int[] members = cluster.getMemberIndexes();

            // For each member, have to ensure it's in the nearest cluster.
            //
            for (int j = 0; j < members.length; j++) {
                int member = members[j];
                double[] coord = mCoordinates[member];
                double minDist = Double.MAX_VALUE;
                int closestClusterIndex = -1;
                for (int k = 0; k < numClusters; k++) {
                    double dist = BasicKMeans.distance(coord, mClusters[k]
                            .getCenter());
                    if (dist < minDist) {
                        minDist = dist;
                        closestClusterIndex = k;
                    }
                } // for (int k...

                // Got a switch.
                if (closestClusterIndex != i) {
                    Integer memberInt = new Integer(member);
                    Set removalSet = (Set) removalMap.get(cluster);
                    if (removalSet == null) {
                        removalSet = new HashSet();
                        removalMap.put(cluster, removalSet);
                    }
                    removalSet.add(memberInt);
                    Set additionSet = (Set) additionMap
                            .get(mClusters[closestClusterIndex]);
                    if (additionSet == null) {
                        additionSet = new HashSet();
                        additionMap.put(mClusters[closestClusterIndex],
                                additionSet);
                    }
                    additionSet.add(memberInt);
                }

            } // for (int j....
        } // for (int i...

        // If the size of removalMap is 0, so is additionMap.
        if (removalMap.size() > 0) {
            int nullCount = 0;
            // Now have to update the cluster whose membership has changed.
            for (int i = 0; i < numClusters; i++) {
                Cluster cluster = mClusters[i];
                Set removalSet = (Set) removalMap.get(cluster);
                Set additionSet = (Set) additionMap.get(cluster);
                if (removalSet != null || additionSet != null) {
                    // If it's in the unsplittables, remove it since it's being updated (replaced).
                    mUnsplittables.remove(cluster);
                    int toRemove = removalSet != null ? removalSet.size() : 0;
                    int toAdd = additionSet != null ? additionSet.size() : 0;
                    int newSize = cluster.getSize() + toAdd - toRemove;

                    if (newSize > 0) {
                        int[] newMembers = new int[newSize];
                        double[] newCenter = new double[dim];
                        int count = 0;
                        int[] oldMembers = cluster.getMemberIndexes();
                        for (int j = 0; j < oldMembers.length; j++) {
                            int member = oldMembers[j];
                            if (removalSet == null
                                    || !removalSet
                                            .contains(new Integer(member))) {
                                newMembers[count++] = member;
                                double[] coordinate = mCoordinates[member];
                                for (int k = 0; k < dim; k++) {
                                    newCenter[k] += coordinate[k];
                                }
                            }
                        } // for (int j...
                        if (additionSet != null) {
                            Iterator it = additionSet.iterator();
                            while (it.hasNext()) {
                                int member = ((Integer) it.next()).intValue();
                                newMembers[count++] = member;
                                double[] coordinate = mCoordinates[member];
                                for (int k = 0; k < dim; k++) {
                                    newCenter[k] += coordinate[k];
                                }
                            }
                        }

                        for (int k = 0; k < dim; k++) {
                            newCenter[k] /= count;
                        }

                        Arrays.sort(newMembers);
                        mClusters[i] = new Cluster(newMembers, newCenter);

                    } else { // newSize == 0
                        mClusters[i] = null;
                        nullCount++;
                    }
                }
            }

            if (nullCount > 0) {
                numClusters -= nullCount;
                Cluster[] newClusters = new Cluster[numClusters];
                int count = 0;
                for (int i = 0; i < mClusters.length; i++) {
                    if (mClusters[i] != null) {
                        newClusters[count++] = mClusters[i];
                    }
                }
                mClusters = newClusters;
            }
        }
    }

    public static double[] computeVariance(double[][] coordinates,
            Cluster cluster) {

        double[] center = cluster.getCenter();
        int[] ids = cluster.getMemberIndexes();

        int dim = center.length;
        double[] variance = new double[dim];
        int n = ids.length;
        if (n > 0) {
            double[] sum = new double[dim];
            for (int i = 0; i < n; i++) {
                double[] coordinate = coordinates[ids[i]];
                for (int j = 0; j < dim; j++) {
                    double d = coordinate[j];
                    sum[j] += d;
                    variance[j] += d * d;
                }
            }
            for (int j = 0; j < dim; j++) {
                variance[j] = Math.max(0.0, (variance[j] - center[j] * sum[j])
                        / n);
            }
        }

        return variance;
    }

    public static double norm1(double[] coordBuffer) {
        double sum = 0.0;
        int n = coordBuffer.length;
        for (int i = 0; i < n; i++) {
            sum += Math.abs(coordBuffer[i]);
        }
        return sum;
    }

    public static double computeBIC(double[][] coordinates, Cluster[] clusters) {

        double bic = 0.0;
        int K = clusters.length;

        if (K > 0 && coordinates.length > 0) {

            // Get the total number of coordinates in the clusters.
            // Don't assume that it's the same as the number of coordinates
            // in the coordinate set.  The cluster set might contain a subset
            // of the coordinates.
            int R = 0;
            for (int i = 0; i < K; i++) {
                R += clusters[i].getSize();
            }

            // Get the dimensionality
            int M = coordinates[0].length;

            double LSum = 0;

            // For each cluster
            for (int i = 0; i < K; i++) {
                Cluster cluster = clusters[i];
                int R_n = cluster.getSize();
                // Estimate variance
                double sigma2 = R_n
                        * norm1(computeVariance(coordinates, cluster));
                if (sigma2 > 0) {
                    sigma2 /= (R_n - K);
                }

                // Estimate log-likelihood
                double L = -R_n / 2 * LOG2PI - (R_n * M / 2) * Math.log(sigma2)
                        - (R_n - K) / 2 + R_n * Math.log(R_n) - R_n
                        * Math.log(R);

                LSum += L;
            }

            // Count the parameters in the model
            double p = K * (M + 1);
            // Compute the criterion
            bic = LSum - p / 2 * Math.log(R);

            // Added this on 3/13/2006 to normalize on cluster size. I don't think
            // the paper we got the bic formula from does this. -- R.Scarberry
            if (R > 0) {
                bic /= R;
            }
        }

        return bic;
    }

    public void addClusteringListener(ClusteringListener l) {
        synchronized (mListeners) {
            if (!mListeners.contains(l)) {
                mListeners.add(l);
            }
        }
    }

    public Cluster[] getClusters() {
        return mClusters;
    }

    public void removeClusteringListener(ClusteringListener l) {
        synchronized (mListeners) {
            mListeners.remove(l);
        }
    }

    /**
     * Posts a message to registered KMeansListeners.
     * 
     * @param message
     */
    private void postClusteringMessage(String message) {
        if (mListeners.size() > 0) {
            synchronized (mListeners) {
                int sz = mListeners.size();
                for (int i = 0; i < sz; i++) {
                    ((ClusteringListener) mListeners.get(i)).clusteringMessage(
                            this, message);
                }
            }
        }
    }

    /**
     * Notifies registered listeners that k-means is complete.
     * 
     * @param clusters the output of clustering.
     * @param executionTime the number of milliseconds taken to cluster.
     */
    private void postClusteringComplete(Cluster[] clusters, long executionTime) {
        if (mListeners.size() > 0) {
            synchronized (mListeners) {
                int sz = mListeners.size();
                for (int i = 0; i < sz; i++) {
                    ((ClusteringListener) mListeners.get(i))
                            .clusteringComplete(this, clusters, executionTime);
                }
            }
        }
    }

    /**
     * Notifies registered listeners that k-means has failed because of
     * a Throwable caught in the run method.
     * 
     * @param err 
     */
    private void postClusteringError(Throwable err) {
        if (mListeners.size() > 0) {
            synchronized (mListeners) {
                int sz = mListeners.size();
                for (int i = 0; i < sz; i++) {
                    ((ClusteringListener) mListeners.get(i)).clusteringError(
                            this, err);
                }
            }
        }
    }

    private void cleanup() {
        mUnsplittables = null;
    }

    public void run() {

        try {

            long startTime = System.currentTimeMillis();

            postClusteringMessage("X-Means clustering started");

            int splits = 0;
            int loop = 0;

            int numCoords = mCoordinates.length;

            if (numCoords > 0) {

                int dim = mCoordinates[0].length;

                int[] initialMembers = new int[numCoords];
                double[] initialCenter = new double[dim];

                for (int i = 0; i < numCoords; i++) {
                    initialMembers[i] = i;
                    double[] coord = mCoordinates[i];
                    for (int j = 0; j < dim; j++) {
                        initialCenter[j] += coord[j];
                    }
                }

                for (int i = 0; i < dim; i++) {
                    initialCenter[i] /= numCoords;
                }

                mClusters = new Cluster[] { new Cluster(initialMembers,
                        initialCenter) };

                int[] splitsToTryA = new int[] { 2, 3, 5, 7, 11 };
                int[] splitsToTryB = new int[] { 2, 3 };

                do {
                    int[] splitsToTry = mClusters.length <= 3 ? splitsToTryA
                            : splitsToTryB;
                    splits = splitClusters(splitsToTry);
                    loop++;
                    postClusteringMessage("... iteration " + loop
                            + " splits = " + splits);
                } while (splits > 0);

                // Because of the splitting process, coordinates can end up in one cluster, even though they're
                // closer to the center of another.  This step fixes that.
                ensureCoordinatesInNearestCluster();

                long executionTime = System.currentTimeMillis() - startTime;

                postClusteringComplete(mClusters, executionTime);

            }

        } catch (Throwable t) {

            t.printStackTrace();

            postClusteringError(t);

        } finally {

            cleanup();

        }
    }
}
